@@ -641,6 +641,9 @@ thread_val {
   
   gcMARK(pr->blocker);
   gcMARK(pr->overflow);
+
+  gcMARK(pr->return_marks_to);
+  gcMARK(pr->returned_marks);
   
   gcMARK(pr->current_local_env);
   gcMARK(pr->current_local_mark);
@@ -651,6 +654,8 @@ thread_val {
   gcMARK(pr->current_local_bindings);
 
   gcMARK(pr->current_mt);
+
+  gcMARK(pr->constant_folding);
   
   gcMARK(pr->overflow_reply);
 
@@ -818,6 +823,7 @@ namespace_val {
   gcMARK(e->insp);
 
   gcMARK(e->rename_set);
+  gcMARK(e->temp_marked_names);
 
   gcMARK(e->syntax);
   gcMARK(e->exp_env);
@@ -941,6 +947,8 @@ module_val {
 
   gcMARK(m->insp);
 
+  gcMARK(m->lang_info);
+
   gcMARK(m->hints);
   gcMARK(m->ii_src);
 
@@ -1047,10 +1055,42 @@ mark_pipe {
   gcBYTES_TO_WORDS(sizeof(Scheme_Pipe));
 }
 
+mark_logger {
+ mark:
+  Scheme_Logger *l = (Scheme_Logger *)p;
+  gcMARK(l->name);
+  gcMARK(l->parent);
+  gcMARK(l->readers);
+  gcMARK(l->timestamp);
+ size:
+  gcBYTES_TO_WORDS(sizeof(Scheme_Logger));
+}
+
+mark_log_reader {
+ mark:
+  Scheme_Log_Reader *lr = (Scheme_Log_Reader *)p;
+  gcMARK(lr->sema);
+  gcMARK(lr->head);
+  gcMARK(lr->tail);
+ size:
+  gcBYTES_TO_WORDS(sizeof(Scheme_Log_Reader));
+}
+
 END type;
 
 /**********************************************************************/
 
+START engine;
+
+engine_val {
+ mark:
+  Scheme_Engine *en = (Scheme_Engine *)p;
+ size:
+  gcBYTES_TO_WORDS(sizeof(Scheme_Engine));
+}
+
+END engine;
+
 START env;
 
 mark_comp_env {
@@ -1111,6 +1151,7 @@ mark_optimize_info {
   gcMARK(i->top_level_consts);
   gcMARK(i->transitive_use);
   gcMARK(i->transitive_use_len);
+  gcMARK(i->context);
 
  size:
   gcBYTES_TO_WORDS(sizeof(Optimize_Info));
@@ -1275,6 +1316,19 @@ END hash;
 
 /**********************************************************************/
 
+START places;
+
+place_val {
+ mark:
+  Scheme_Place *pr = (Scheme_Place *)p;
+ size:
+  gcBYTES_TO_WORDS(sizeof(Scheme_Place));
+}
+
+END places;
+
+/**********************************************************************/
+
 START portfun;
 
 mark_load_handler_data {
@@ -1676,6 +1730,7 @@ mark_syncing {
   gcMARK(w->wrapss);
   gcMARK(w->nackss);
   gcMARK(w->reposts);
+  gcMARK(w->accepts);
   gcMARK(w->disable_break);
 
  size:
@@ -1839,6 +1894,7 @@ mark_struct_property {
   Scheme_Struct_Property *i = (Scheme_Struct_Property *)p;
   gcMARK(i->name);
   gcMARK(i->guard);
+  gcMARK(i->supers);
  size:
   gcBYTES_TO_WORDS(sizeof(Scheme_Struct_Property));
 }
@@ -2110,8 +2166,10 @@ native_unclosed_proc {
   int i;
 
   gcMARK(d->u2.name);
-  for (i = d->retain_count; i--; ) {
-    gcMARK(d->retained[i]);
+  if (d->retained) {
+    for (i = SCHEME_INT_VAL(d->retained[0]); i--; ) {
+      gcMARK(d->retained[i]);
+    }
   }
   if (d->closure_size < 0) {
     gcMARK(d->u.arities);
