@@ -68,7 +68,9 @@ MZ_EXTERN Scheme_Object *scheme_current_break_cell();
 /*========================================================================*/
 
 #ifndef LINK_EXTENSIONS_BY_TABLE
-MZ_EXTERN Scheme_Thread *scheme_current_thread;
+# ifndef MZ_USE_PLACES
+MZ_EXTERN THREAD_LOCAL Scheme_Thread *scheme_current_thread;
+# endif
 MZ_EXTERN volatile int scheme_fuel_counter;
 #else
 MZ_EXTERN Scheme_Thread **scheme_current_thread_ptr;
@@ -176,6 +178,13 @@ MZ_EXTERN void scheme_warning(char *msg, ...);
 
 MZ_EXTERN void scheme_raise(Scheme_Object *exn);
 
+MZ_EXTERN int scheme_log_level_p(Scheme_Logger *logger, int level);
+MZ_EXTERN void scheme_log(Scheme_Logger *logger, int level, int flags,
+                          char *msg, ...);
+MZ_EXTERN void scheme_log_message(Scheme_Logger *logger, int level, char *buffer, long len, Scheme_Object *data);
+MZ_EXTERN void scheme_log_abort(char *buffer);
+MZ_EXTERN void scheme_out_of_memory_abort();
+
 MZ_EXTERN void scheme_wrong_count(const char *name, int minc, int maxc,
 				  int argc, Scheme_Object **argv);
 MZ_EXTERN void scheme_wrong_count_m(const char *name, int minc, int maxc,
@@ -215,15 +224,15 @@ MZ_EXTERN char *scheme_get_type_name(Scheme_Type type);
 /*========================================================================*/
 
 MZ_EXTERN Scheme_Object scheme_eof[1];
-MZ_EXTERN Scheme_Object *scheme_make_eof(void);
+XFORM_NONGCING MZ_EXTERN Scheme_Object *scheme_make_eof(void);
 MZ_EXTERN Scheme_Object scheme_null[1];
-MZ_EXTERN Scheme_Object *scheme_make_null(void);
+XFORM_NONGCING MZ_EXTERN Scheme_Object *scheme_make_null(void);
 MZ_EXTERN Scheme_Object scheme_true[1];
-MZ_EXTERN Scheme_Object *scheme_make_true(void);
+XFORM_NONGCING MZ_EXTERN Scheme_Object *scheme_make_true(void);
 MZ_EXTERN Scheme_Object scheme_false[1];
-MZ_EXTERN Scheme_Object *scheme_make_false(void);
+XFORM_NONGCING MZ_EXTERN Scheme_Object *scheme_make_false(void);
 MZ_EXTERN Scheme_Object scheme_void[1];
-MZ_EXTERN Scheme_Object *scheme_make_void(void);
+XFORM_NONGCING MZ_EXTERN Scheme_Object *scheme_make_void(void);
 MZ_EXTERN Scheme_Object scheme_undefined[1];
 MZ_EXTERN Scheme_Object *scheme_tail_call_waiting;
 MZ_EXTERN Scheme_Object *scheme_multiple_values;
@@ -351,6 +360,13 @@ MZ_EXTERN void *GC_malloc_uncollectable(size_t size_in_bytes);
 # endif
 #endif
 
+MZ_EXTERN void *scheme_malloc_code(long size);
+MZ_EXTERN void scheme_free_code(void *p);
+#ifndef MZ_PRECISE_GC
+MZ_EXTERN void *scheme_malloc_gcable_code(long size);
+#endif
+
+
 MZ_EXTERN void *scheme_malloc_eternal(size_t n);
 MZ_EXTERN void scheme_end_stubborn_change(void *p);
 
@@ -383,12 +399,13 @@ MZ_EXTERN void scheme_gc_ptr_ok(void *p);
 MZ_EXTERN void scheme_collect_garbage(void);
 
 #ifdef MZ_PRECISE_GC
-MZ_EXTERN void **GC_variable_stack;
+MZ_EXTERN THREAD_LOCAL void **GC_variable_stack;
 MZ_EXTERN void GC_register_traversers(short tag, Size_Proc size, Mark_Proc mark, Fixup_Proc fixup,
 				      int is_constant_size, int is_atomic);
 MZ_EXTERN void *GC_resolve(void *p);
 MZ_EXTERN void GC_mark(const void *p);
 MZ_EXTERN void GC_fixup(void *p);
+MZ_EXTERN void *GC_fixup_self(void *p);
 #endif
 
 MZ_EXTERN void **scheme_malloc_immobile_box(void *p);
@@ -408,6 +425,7 @@ MZ_EXTERN Scheme_Bucket_Table *scheme_clone_bucket_table(Scheme_Bucket_Table *bt
 
 MZ_EXTERN Scheme_Hash_Table *scheme_make_hash_table(int type);
 MZ_EXTERN Scheme_Hash_Table *scheme_make_hash_table_equal();
+MZ_EXTERN Scheme_Hash_Table *scheme_make_hash_table_eqv();
 MZ_EXTERN void scheme_hash_set(Scheme_Hash_Table *table, Scheme_Object *key, Scheme_Object *val);
 MZ_EXTERN Scheme_Object *scheme_hash_get(Scheme_Hash_Table *table, Scheme_Object *key);
 XFORM_NONGCING MZ_EXTERN Scheme_Object *scheme_eq_hash_get(Scheme_Hash_Table *table, Scheme_Object *key);
@@ -415,15 +433,17 @@ MZ_EXTERN void scheme_hash_set_atomic(Scheme_Hash_Table *table, Scheme_Object *k
 MZ_EXTERN Scheme_Object *scheme_hash_get_atomic(Scheme_Hash_Table *table, Scheme_Object *key);
 MZ_EXTERN int scheme_hash_table_equal(Scheme_Hash_Table *t1, Scheme_Hash_Table *t2);
 MZ_EXTERN int scheme_is_hash_table_equal(Scheme_Object *o);
+MZ_EXTERN int scheme_is_hash_table_eqv(Scheme_Object *o);
 MZ_EXTERN Scheme_Hash_Table *scheme_clone_hash_table(Scheme_Hash_Table *bt);
 
-MZ_EXTERN Scheme_Hash_Tree *scheme_make_hash_tree(int eql);
+MZ_EXTERN Scheme_Hash_Tree *scheme_make_hash_tree(int kind);
 MZ_EXTERN Scheme_Hash_Tree *scheme_hash_tree_set(Scheme_Hash_Tree *tree, Scheme_Object *key, Scheme_Object *val);
 MZ_EXTERN Scheme_Object *scheme_hash_tree_get(Scheme_Hash_Tree *tree, Scheme_Object *key);
 MZ_EXTERN long scheme_hash_tree_next(Scheme_Hash_Tree *tree, long pos);
 MZ_EXTERN int scheme_hash_tree_index(Scheme_Hash_Tree *tree, long pos, Scheme_Object **_key, Scheme_Object **_val);
 MZ_EXTERN int scheme_hash_tree_equal(Scheme_Hash_Tree *t1, Scheme_Hash_Tree *t2);
 MZ_EXTERN int scheme_is_hash_tree_equal(Scheme_Object *o);
+MZ_EXTERN int scheme_is_hash_tree_eqv(Scheme_Object *o);
 
 /*========================================================================*/
 /*                   basic Scheme value constructors                      */
@@ -791,6 +811,7 @@ MZ_EXTERN Scheme_Object *scheme_write_special_evt_via_write_special(Scheme_Outpu
 
 MZ_EXTERN Scheme_Object *scheme_open_input_file(const char *name, const char *who);
 MZ_EXTERN Scheme_Object *scheme_open_output_file(const char *name, const char *who);
+MZ_EXTERN Scheme_Object *scheme_open_input_output_file(const char *name, const char *who, Scheme_Object **oport);
 MZ_EXTERN Scheme_Object *scheme_open_output_file_with_mode(const char *name, const char *who, int text);
 
 MZ_EXTERN Scheme_Object *scheme_make_file_input_port(FILE *fp);
@@ -882,8 +903,6 @@ MZ_EXTERN Scheme_Object *scheme_read_byte_string(Scheme_Object *port);
 MZ_EXTERN Scheme_Object *scheme_make_namespace(int argc, Scheme_Object *argv[]);
 MZ_EXTERN void scheme_add_namespace_option(Scheme_Object *key, void (*f)(Scheme_Env *));
 
-MZ_EXTERN void scheme_require_from_original_env(Scheme_Env *env, int syntax_only);
-
 MZ_EXTERN void scheme_add_global(const char *name, Scheme_Object *val, Scheme_Env *env);
 MZ_EXTERN void scheme_add_global_symbol(Scheme_Object *name, Scheme_Object *val,
 			      Scheme_Env *env);
@@ -997,6 +1016,8 @@ MZ_EXTERN long scheme_equal_hash_key(Scheme_Object *o);
 MZ_EXTERN long scheme_equal_hash_key2(Scheme_Object *o);
 MZ_EXTERN long scheme_recur_equal_hash_key(Scheme_Object *o, void *cycle_data);
 MZ_EXTERN long scheme_recur_equal_hash_key2(Scheme_Object *o, void *cycle_data);
+MZ_EXTERN long scheme_eqv_hash_key(Scheme_Object *o);
+MZ_EXTERN long scheme_eqv_hash_key2(Scheme_Object *o);
 
 MZ_EXTERN void scheme_set_type_equality(Scheme_Type type, 
                                         Scheme_Equal_Proc f,
@@ -1057,8 +1078,6 @@ MZ_EXTERN int scheme_check_proc_arity2(const char *where, int a,
 MZ_EXTERN char *scheme_make_provided_string(Scheme_Object *o, int count, int *len);
 MZ_EXTERN char *scheme_make_args_string(char *s, int which, int argc, Scheme_Object **argv, long *len);
 
-MZ_EXTERN void scheme_no_dumps(char *why);
-
 MZ_EXTERN const char *scheme_system_library_subpath();
 
 MZ_EXTERN void scheme_signal_received(void);
