@@ -54,7 +54,9 @@ Scheme_Object *(*scheme_current_break_cell)();
 /*                                threads                                 */
 /*========================================================================*/
 #ifndef LINK_EXTENSIONS_BY_TABLE
+# ifndef MZ_USE_PLACES
 Scheme_Thread *scheme_current_thread;
+# endif
 volatile int scheme_fuel_counter;
 #else
 Scheme_Thread **scheme_current_thread_ptr;
@@ -138,6 +140,12 @@ void (*scheme_signal_error)(const char *msg, ...);
 void (*scheme_raise_exn)(int exnid, ...);
 void (*scheme_warning)(char *msg, ...);
 void (*scheme_raise)(Scheme_Object *exn);
+int (*scheme_log_level_p)(Scheme_Logger *logger, int level);
+void (*scheme_log)(Scheme_Logger *logger, int level, int flags,
+                          char *msg, ...);
+void (*scheme_log_message)(Scheme_Logger *logger, int level, char *buffer, long len, Scheme_Object *data);
+void (*scheme_log_abort)(char *buffer);
+void (*scheme_out_of_memory_abort)();
 void (*scheme_wrong_count)(const char *name, int minc, int maxc,
 				  int argc, Scheme_Object **argv);
 void (*scheme_wrong_count_m)(const char *name, int minc, int maxc,
@@ -285,6 +293,11 @@ void *(*GC_malloc_uncollectable)(size_t size_in_bytes);
 #  endif
 # endif
 #endif
+void *(*scheme_malloc_code)(long size);
+void (*scheme_free_code)(void *p);
+#ifndef MZ_PRECISE_GC
+void *(*scheme_malloc_gcable_code)(long size);
+#endif
 void *(*scheme_malloc_eternal)(size_t n);
 void (*scheme_end_stubborn_change)(void *p);
 void *(*scheme_calloc)(size_t num, size_t size);
@@ -316,6 +329,7 @@ void (*GC_register_traversers)(short tag, Size_Proc size, Mark_Proc mark, Fixup_
 void *(*GC_resolve)(void *p);
 void (*GC_mark)(const void *p);
 void (*GC_fixup)(void *p);
+void *(*GC_fixup_self)(void *p);
 #endif
 void **(*scheme_malloc_immobile_box)(void *p);
 void (*scheme_free_immobile_box)(void **b);
@@ -331,6 +345,7 @@ int (*scheme_bucket_table_equal)(Scheme_Bucket_Table *t1, Scheme_Bucket_Table *t
 Scheme_Bucket_Table *(*scheme_clone_bucket_table)(Scheme_Bucket_Table *bt);
 Scheme_Hash_Table *(*scheme_make_hash_table)(int type);
 Scheme_Hash_Table *(*scheme_make_hash_table_equal)();
+Scheme_Hash_Table *(*scheme_make_hash_table_eqv)();
 void (*scheme_hash_set)(Scheme_Hash_Table *table, Scheme_Object *key, Scheme_Object *val);
 Scheme_Object *(*scheme_hash_get)(Scheme_Hash_Table *table, Scheme_Object *key);
 Scheme_Object *(*scheme_eq_hash_get)(Scheme_Hash_Table *table, Scheme_Object *key);
@@ -338,14 +353,16 @@ void (*scheme_hash_set_atomic)(Scheme_Hash_Table *table, Scheme_Object *key, Sch
 Scheme_Object *(*scheme_hash_get_atomic)(Scheme_Hash_Table *table, Scheme_Object *key);
 int (*scheme_hash_table_equal)(Scheme_Hash_Table *t1, Scheme_Hash_Table *t2);
 int (*scheme_is_hash_table_equal)(Scheme_Object *o);
+int (*scheme_is_hash_table_eqv)(Scheme_Object *o);
 Scheme_Hash_Table *(*scheme_clone_hash_table)(Scheme_Hash_Table *bt);
-Scheme_Hash_Tree *(*scheme_make_hash_tree)(int eql);
+Scheme_Hash_Tree *(*scheme_make_hash_tree)(int kind);
 Scheme_Hash_Tree *(*scheme_hash_tree_set)(Scheme_Hash_Tree *tree, Scheme_Object *key, Scheme_Object *val);
 Scheme_Object *(*scheme_hash_tree_get)(Scheme_Hash_Tree *tree, Scheme_Object *key);
 long (*scheme_hash_tree_next)(Scheme_Hash_Tree *tree, long pos);
 int (*scheme_hash_tree_index)(Scheme_Hash_Tree *tree, long pos, Scheme_Object **_key, Scheme_Object **_val);
 int (*scheme_hash_tree_equal)(Scheme_Hash_Tree *t1, Scheme_Hash_Tree *t2);
 int (*scheme_is_hash_tree_equal)(Scheme_Object *o);
+int (*scheme_is_hash_tree_eqv)(Scheme_Object *o);
 /*========================================================================*/
 /*                   basic Scheme value constructors                      */
 /*========================================================================*/
@@ -667,6 +684,7 @@ Scheme_Object *(*scheme_write_special_evt_via_write_special)(Scheme_Output_Port
 								    Scheme_Object *special);
 Scheme_Object *(*scheme_open_input_file)(const char *name, const char *who);
 Scheme_Object *(*scheme_open_output_file)(const char *name, const char *who);
+Scheme_Object *(*scheme_open_input_output_file)(const char *name, const char *who, Scheme_Object **oport);
 Scheme_Object *(*scheme_open_output_file_with_mode)(const char *name, const char *who, int text);
 Scheme_Object *(*scheme_make_file_input_port)(FILE *fp);
 Scheme_Object *(*scheme_make_named_file_input_port)(FILE *fp, Scheme_Object *name);
@@ -738,7 +756,6 @@ Scheme_Object *(*scheme_read_byte_string)(Scheme_Object *port);
 /*========================================================================*/
 Scheme_Object *(*scheme_make_namespace)(int argc, Scheme_Object *argv[]);
 void (*scheme_add_namespace_option)(Scheme_Object *key, void (*f)(Scheme_Env *));
-void (*scheme_require_from_original_env)(Scheme_Env *env, int syntax_only);
 void (*scheme_add_global)(const char *name, Scheme_Object *val, Scheme_Env *env);
 void (*scheme_add_global_symbol)(Scheme_Object *name, Scheme_Object *val,
 			      Scheme_Env *env);
@@ -827,6 +844,8 @@ long (*scheme_equal_hash_key)(Scheme_Object *o);
 long (*scheme_equal_hash_key2)(Scheme_Object *o);
 long (*scheme_recur_equal_hash_key)(Scheme_Object *o, void *cycle_data);
 long (*scheme_recur_equal_hash_key2)(Scheme_Object *o, void *cycle_data);
+long (*scheme_eqv_hash_key)(Scheme_Object *o);
+long (*scheme_eqv_hash_key2)(Scheme_Object *o);
 void (*scheme_set_type_equality)(Scheme_Type type, 
                                         Scheme_Equal_Proc f,
                                         Scheme_Primary_Hash_Proc hash1,
@@ -871,7 +890,6 @@ int (*scheme_check_proc_arity2)(const char *where, int a,
 				       int false_ok);
 char *(*scheme_make_provided_string)(Scheme_Object *o, int count, int *len);
 char *(*scheme_make_args_string)(char *s, int which, int argc, Scheme_Object **argv, long *len);
-void (*scheme_no_dumps)(char *why);
 const char *(*scheme_system_library_subpath)();
 void (*scheme_signal_received)(void);
 int (*scheme_char_strlen)(const mzchar *s);
