@@ -19,7 +19,9 @@
 #define scheme_inherit_cells (scheme_extension_table->scheme_inherit_cells)
 #define scheme_current_break_cell (scheme_extension_table->scheme_current_break_cell)
 #ifndef LINK_EXTENSIONS_BY_TABLE
+# ifndef MZ_USE_PLACES
 #define scheme_current_thread (scheme_extension_table->scheme_current_thread)
+# endif
 #define scheme_fuel_counter (scheme_extension_table->scheme_fuel_counter)
 #else
 #define scheme_current_thread_ptr (scheme_extension_table->scheme_current_thread_ptr)
@@ -83,6 +85,11 @@
 #define scheme_raise_exn (scheme_extension_table->scheme_raise_exn)
 #define scheme_warning (scheme_extension_table->scheme_warning)
 #define scheme_raise (scheme_extension_table->scheme_raise)
+#define scheme_log_level_p (scheme_extension_table->scheme_log_level_p)
+#define scheme_log (scheme_extension_table->scheme_log)
+#define scheme_log_message (scheme_extension_table->scheme_log_message)
+#define scheme_log_abort (scheme_extension_table->scheme_log_abort)
+#define scheme_out_of_memory_abort (scheme_extension_table->scheme_out_of_memory_abort)
 #define scheme_wrong_count (scheme_extension_table->scheme_wrong_count)
 #define scheme_wrong_count_m (scheme_extension_table->scheme_wrong_count_m)
 #define scheme_case_lambda_wrong_count (scheme_extension_table->scheme_case_lambda_wrong_count)
@@ -189,6 +196,11 @@
 #  endif
 # endif
 #endif
+#define scheme_malloc_code (scheme_extension_table->scheme_malloc_code)
+#define scheme_free_code (scheme_extension_table->scheme_free_code)
+#ifndef MZ_PRECISE_GC
+#define scheme_malloc_gcable_code (scheme_extension_table->scheme_malloc_gcable_code)
+#endif
 #define scheme_malloc_eternal (scheme_extension_table->scheme_malloc_eternal)
 #define scheme_end_stubborn_change (scheme_extension_table->scheme_end_stubborn_change)
 #define scheme_calloc (scheme_extension_table->scheme_calloc)
@@ -216,6 +228,7 @@
 #define GC_resolve (scheme_extension_table->GC_resolve)
 #define GC_mark (scheme_extension_table->GC_mark)
 #define GC_fixup (scheme_extension_table->GC_fixup)
+#define GC_fixup_self (scheme_extension_table->GC_fixup_self)
 #endif
 #define scheme_malloc_immobile_box (scheme_extension_table->scheme_malloc_immobile_box)
 #define scheme_free_immobile_box (scheme_extension_table->scheme_free_immobile_box)
@@ -228,6 +241,7 @@
 #define scheme_clone_bucket_table (scheme_extension_table->scheme_clone_bucket_table)
 #define scheme_make_hash_table (scheme_extension_table->scheme_make_hash_table)
 #define scheme_make_hash_table_equal (scheme_extension_table->scheme_make_hash_table_equal)
+#define scheme_make_hash_table_eqv (scheme_extension_table->scheme_make_hash_table_eqv)
 #define scheme_hash_set (scheme_extension_table->scheme_hash_set)
 #define scheme_hash_get (scheme_extension_table->scheme_hash_get)
 #define scheme_eq_hash_get (scheme_extension_table->scheme_eq_hash_get)
@@ -235,6 +249,7 @@
 #define scheme_hash_get_atomic (scheme_extension_table->scheme_hash_get_atomic)
 #define scheme_hash_table_equal (scheme_extension_table->scheme_hash_table_equal)
 #define scheme_is_hash_table_equal (scheme_extension_table->scheme_is_hash_table_equal)
+#define scheme_is_hash_table_eqv (scheme_extension_table->scheme_is_hash_table_eqv)
 #define scheme_clone_hash_table (scheme_extension_table->scheme_clone_hash_table)
 #define scheme_make_hash_tree (scheme_extension_table->scheme_make_hash_tree)
 #define scheme_hash_tree_set (scheme_extension_table->scheme_hash_tree_set)
@@ -243,6 +258,7 @@
 #define scheme_hash_tree_index (scheme_extension_table->scheme_hash_tree_index)
 #define scheme_hash_tree_equal (scheme_extension_table->scheme_hash_tree_equal)
 #define scheme_is_hash_tree_equal (scheme_extension_table->scheme_is_hash_tree_equal)
+#define scheme_is_hash_tree_eqv (scheme_extension_table->scheme_is_hash_tree_eqv)
 #define scheme_make_prim (scheme_extension_table->scheme_make_prim)
 #define scheme_make_noneternal_prim (scheme_extension_table->scheme_make_noneternal_prim)
 #define scheme_make_prim_w_arity (scheme_extension_table->scheme_make_prim_w_arity)
@@ -445,6 +461,7 @@
 #define scheme_write_special_evt_via_write_special (scheme_extension_table->scheme_write_special_evt_via_write_special)
 #define scheme_open_input_file (scheme_extension_table->scheme_open_input_file)
 #define scheme_open_output_file (scheme_extension_table->scheme_open_output_file)
+#define scheme_open_input_output_file (scheme_extension_table->scheme_open_input_output_file)
 #define scheme_open_output_file_with_mode (scheme_extension_table->scheme_open_output_file_with_mode)
 #define scheme_make_file_input_port (scheme_extension_table->scheme_make_file_input_port)
 #define scheme_make_named_file_input_port (scheme_extension_table->scheme_make_named_file_input_port)
@@ -509,7 +526,6 @@
 #define scheme_read_byte_string (scheme_extension_table->scheme_read_byte_string)
 #define scheme_make_namespace (scheme_extension_table->scheme_make_namespace)
 #define scheme_add_namespace_option (scheme_extension_table->scheme_add_namespace_option)
-#define scheme_require_from_original_env (scheme_extension_table->scheme_require_from_original_env)
 #define scheme_add_global (scheme_extension_table->scheme_add_global)
 #define scheme_add_global_symbol (scheme_extension_table->scheme_add_global_symbol)
 #define scheme_make_envunbox (scheme_extension_table->scheme_make_envunbox)
@@ -566,6 +582,8 @@
 #define scheme_equal_hash_key2 (scheme_extension_table->scheme_equal_hash_key2)
 #define scheme_recur_equal_hash_key (scheme_extension_table->scheme_recur_equal_hash_key)
 #define scheme_recur_equal_hash_key2 (scheme_extension_table->scheme_recur_equal_hash_key2)
+#define scheme_eqv_hash_key (scheme_extension_table->scheme_eqv_hash_key)
+#define scheme_eqv_hash_key2 (scheme_extension_table->scheme_eqv_hash_key2)
 #define scheme_set_type_equality (scheme_extension_table->scheme_set_type_equality)
 #define scheme_recur_equal (scheme_extension_table->scheme_recur_equal)
 #define scheme_build_list (scheme_extension_table->scheme_build_list)
@@ -603,7 +621,6 @@
 #define scheme_check_proc_arity2 (scheme_extension_table->scheme_check_proc_arity2)
 #define scheme_make_provided_string (scheme_extension_table->scheme_make_provided_string)
 #define scheme_make_args_string (scheme_extension_table->scheme_make_args_string)
-#define scheme_no_dumps (scheme_extension_table->scheme_no_dumps)
 #define scheme_system_library_subpath (scheme_extension_table->scheme_system_library_subpath)
 #define scheme_signal_received (scheme_extension_table->scheme_signal_received)
 #define scheme_char_strlen (scheme_extension_table->scheme_char_strlen)
