@@ -266,12 +266,17 @@ scheme_init_type (Scheme_Env *env)
 
   set_name(scheme_already_comp_type, "<internal:already-compiled>");
 
+  set_name(scheme_logger_type, "<logger>");
+  set_name(scheme_log_reader_type, "<log-receiver>");
+
   set_name(_scheme_values_types_, "<resurrected>");
   set_name(_scheme_compiled_values_types_, "<internal>");
 
 #ifdef MZ_GC_BACKTRACE
   set_name(scheme_rt_meta_cont, "<meta-continuation>");
 #endif
+  set_name(scheme_place_type, "<place>");
+  set_name(scheme_engine_type, "<engine>");
 }
 
 Scheme_Type scheme_make_type(const char *name)
@@ -610,6 +615,9 @@ void scheme_register_traversers(void)
 
   GC_REG_TRAV(scheme_resolved_module_path_type, small_object);
 
+  GC_REG_TRAV(scheme_logger_type, mark_logger);
+  GC_REG_TRAV(scheme_log_reader_type, mark_log_reader);
+
   GC_REG_TRAV(scheme_rt_runstack, runstack_val);
 }
 
