@@ -130,7 +130,8 @@ static XtResource MenuResources[] =
 	offset(menu.horizontal), XtRImmediate, (XtPointer) True},
     {XtNforChoice, XtCForChoice, XtRBoolean, sizeof(Boolean),
 	offset(menu.forChoice), XtRImmediate, (XtPointer) False},
-
+    {XtNforPopup, XtCForPopup, XtRBoolean, sizeof(Boolean),
+	offset(menu.forPopup), XtRImmediate, (XtPointer) False},
 
     /* menu structure */
     {XtNmenu, XtCMenu, XtRPointer, sizeof(XtPointer),
@@ -450,7 +451,7 @@ static void MenuResize(w)
  *
  *****************************************************************************/
 
-static int HandleMotionEvent(MenuWidget mw, XMotionEvent *ev);
+static int HandleMotionEvent(MenuWidget mw, XMotionEvent *ev, int is_click);
 static void UnhighlightItem(MenuWidget mw, menu_state *ms, menu_item *item);
 static void HighlightItem(MenuWidget mw, menu_state *ms, menu_item *item);
 static void MoveSelection(MenuWidget mw, int direction);
@@ -490,7 +491,9 @@ static void Start(w, ev, params, num_params)
       mw->menu.grabbed = TRUE;
     }
 
-    if (!HandleMotionEvent(mw, &ev->xmotion) && !num_params)
+    if (!HandleMotionEvent(mw, &ev->xmotion, 1)
+        /* *num_params is positive when called to start a popup menu */
+        && !*num_params)
       DoSelect(w, CurrentTime, 1);
 }
 
@@ -507,7 +510,7 @@ static void Drag(w, event, params, num_params)
     int           y     = ev->y_root;
     int           state = ev->state;
 
-    HandleMotionEvent(mw, ev);
+    HandleMotionEvent(mw, ev, 0);
     XSync(XtDisplay(mw), FALSE);
     /* allow motion events to be generated again */
     if ((!ev->is_hint
@@ -517,7 +520,7 @@ static void Drag(w, event, params, num_params)
 			  &ev->x, &ev->y, &ev->state))
 	&& ev->state == state
 	&& (ev->x_root != x || ev->y_root != y)) {
-      HandleMotionEvent(mw, ev);
+      HandleMotionEvent(mw, ev, 0);
       XSync(XtDisplay(mw), FALSE);
     }
 }
@@ -576,7 +579,11 @@ static void Select(w, event, params, num_params)
 
   mw->menu.moused_out = 0;
 
-  force = !HandleMotionEvent(mw, ev);
+  if (!mw->menu.forPopup 
+      && (!mw->menu.state || !mw->menu.state->selected))
+    return;
+
+  force = !HandleMotionEvent(mw, ev, 0);
   if (!force)
     force = mw->menu.moused_out;
   
@@ -1478,10 +1485,10 @@ static void timer_callback(XtPointer client_data, XtIntervalId * timer)
 		&ev.x_root, &ev.y_root,
 		&ev.x, &ev.y, &ev.state);
 
-  HandleMotionEvent(mw, &ev);
+  HandleMotionEvent(mw, &ev, 0);
 }
 
-static int HandleMotionEvent(MenuWidget mw, XMotionEvent *ev)
+static int HandleMotionEvent(MenuWidget mw, XMotionEvent *ev, int is_click)
 {
     menu_state *ms = NULL;
     menu_item  *item = NULL;
@@ -1549,6 +1556,11 @@ static int HandleMotionEvent(MenuWidget mw, XMotionEvent *ev)
     if (!foundone)
       mw->menu.moused_out = !in_extra_region;
 
+    if (!mw->menu.forPopup 
+        && (is_click && ms && (item == ms->selected))) { /* pointer on the same item and a click */
+      return 0;
+    }
+
     if (!item) { /* if pointer not on menu_item unhighlight last selected */
       UnhighlightItem(mw, mw->menu.state, mw->menu.state->selected);
       if (scroll) {
@@ -1580,8 +1592,11 @@ static int HandleMotionEvent(MenuWidget mw, XMotionEvent *ev)
       }
       return in_extra_region;
     }
-    if (item == ms->selected) /* pointer on the same item */
+
+    if (item == ms->selected) { /* pointer on the same item (and not a click) */
       return 1;
+    }
+
     /* unhighlight old item on same level (ms!) and highlight new item */
     UnhighlightItem(mw, ms, ms->selected);
     HighlightItem(mw, ms, item);
@@ -1722,7 +1737,7 @@ void Xaw3dPopupMenuAtPos(MenuWidget mw, int x, int y)
     mw->menu.state->y = y+border_width;
     /* init first motion event */
     ev.x_root = x; ev.y_root = y;
-    HandleMotionEvent(mw, (XMotionEvent*)&ev);
+    HandleMotionEvent(mw, (XMotionEvent*)&ev, 0);
 }
 
 int xwMenuIsPoppedUp(Widget w)
